/*
 * Copyright (c) 2004 Thorsten Kukuk <kukuk@suse.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, and the entire permission notice in its entirety,
 *    including the disclaimer of warranties.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior
 *    written permission.
 *
 * ALTERNATIVELY, this product may be distributed under the terms of
 * the GNU Public License, in which case the provisions of the GPL are
 * required INSTEAD OF the above restrictions.  (This clause is
 * necessary due to a potential bad interaction between the GPL and
 * the restrictions contained in a BSD-style copyright.)
*
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#if defined(HAVE_CONFIG_H)
#include "config.h"
#endif

#define _GNU_SOURCE

#include <iconv.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <langinfo.h>

#include "utf8conv.h"

static char *
call_iconv (const char *from, const char *to, const char *str)
{
  size_t nconv, srclen, dstlen;
  iconv_t ih = iconv_open (to, from);
  char *srcstr, *dststr, *resstr;

  if (ih == (iconv_t) -1)
    {
      /* Something went wrong.  */
      if (errno == EINVAL)
	fprintf (stderr, "conversion from `%s' to `%s' not available.\n",
		 from, to);
      else
	perror ("iconv_open");

      return NULL;
    }

  srcstr = strdupa (str);
  srclen = strlen (srcstr);
  dstlen = srclen * 4;
  dststr = malloc (dstlen);
  if (dststr == NULL)
    return NULL;
  resstr = dststr;

  nconv = iconv (ih, &srcstr, &srclen, &dststr, &dstlen);
  if (nconv == (size_t) -1)
    {
      if (errno != EILSEQ)
	perror ("iconv");
      free (resstr);
      return strdup (str);
    }
  *dststr = '\0';

  if (iconv_close (ih) != 0)
    perror ("iconv_close");

  return resstr;
}

char *
utf8_to_locale (const char *str)
{
  char *to = nl_langinfo (CODESET);
  char *res;

  if (to == NULL ||
      strcmp (to, "C") == 0 ||
      strcmp (to, "UTF-8") == 0)
    return strdup (str);

  res = call_iconv ("UTF-8", to, str);

  if (res == NULL)
    return strdup (str);

  return res;
}

char *
locale_to_utf8 (const char *str)
{
  char *from = nl_langinfo (CODESET);
  char *res;

  if (from == NULL ||
      strcmp (from, "C") == 0 ||
      strcmp (from, "UTF-8") == 0)
    return strdup (str);

  res = call_iconv (from, "UTF-8", str);

  if (res == NULL)
    return strdup (str);

  return res;
}
